// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=* 
// ** Copyright UCAR (c) 1990 - 2016                                         
// ** University Corporation for Atmospheric Research (UCAR)                 
// ** National Center for Atmospheric Research (NCAR)                        
// ** Boulder, Colorado, USA                                                 
// ** BSD licence applies - redistribution and use in source and binary      
// ** forms, with or without modification, are permitted provided that       
// ** the following conditions are met:                                      
// ** 1) If the software is modified to produce derivative works,            
// ** such modified software should be clearly marked, so as not             
// ** to confuse it with the version available from UCAR.                    
// ** 2) Redistributions of source code must retain the above copyright      
// ** notice, this list of conditions and the following disclaimer.          
// ** 3) Redistributions in binary form must reproduce the above copyright   
// ** notice, this list of conditions and the following disclaimer in the    
// ** documentation and/or other materials provided with the distribution.   
// ** 4) Neither the name of UCAR nor the names of its contributors,         
// ** if any, may be used to endorse or promote products derived from        
// ** this software without specific prior written permission.               
// ** DISCLAIMER: THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS  
// ** OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED      
// ** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.    
// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=* 
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/

// RCS info
//   $Author: dixon $
//   $Locker:  $
//   $Date: 2016/03/03 18:19:27 $
//   $Id: GridTemplate.cc,v 1.14 2016/03/03 18:19:27 dixon Exp $
//   $Revision: 1.14 $
//   $State: Exp $
 
/**-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-**/
/*********************************************************************
 * GridTemplate.cc: class implementing a template to be applied to
 *                  gridded data.
 *
 * RAP, NCAR, Boulder CO
 *
 * January 1999
 *
 * Nancy Rehak
 *
 *********************************************************************/

#include <vector>

#include <math.h>
#include <cstdio>

#include "vx_log.h"

#include "GridTemplate.h"
#include "GridOffset.h"

#include "RectangularTemplate.h"
#include "CircularTemplate.h"

using namespace std;

/**********************************************************************
 * Constructors
 */

GridTemplate::GridTemplate(void)
{
  // Do nothing
}


GridTemplate::GridTemplate(const GridTemplate& rhs)
{
	vector< GridOffset* >::const_iterator offset_iter;
	
	for (offset_iter = rhs._offsetList.begin();
	     offset_iter != rhs._offsetList.end(); ++offset_iter)
		_offsetList.push_back(new GridOffset(*offset_iter));
  
	_pointInGridBase = rhs._pointInGridBase;
	_pointInGridNumX = rhs._pointInGridNumX;
	_pointInGridNumY = rhs._pointInGridNumY;
	_pointInGridReturn = rhs._pointInGridReturn;
  
}	


/**********************************************************************
 * Destructor
 */

GridTemplate::~GridTemplate(void)
{
  // Reclaim the space for the offset list

  vector< GridOffset* >::iterator list_iter;
  for (list_iter = _offsetList.begin(); list_iter != _offsetList.end();
       ++list_iter)
    delete *list_iter;
  
  _offsetList.erase(_offsetList.begin(), _offsetList.end());
  
}
  

/**********************************************************************
 * getFirstInGrid() - Get the first template grid point within the given
 *                    grid.
 *
 * Returns a pointer to a static object which must NOT be deleted by the
 * calling routine.
 */

GridPoint *GridTemplate::getFirstInGrid(const int &base_x, const int &base_y,
					const int &nx, const int &ny) const
{
  // Set up the iterator and save the grid information

  _pointInGridIterator = _offsetList.begin();
  
  _pointInGridBase.x = base_x;
  _pointInGridBase.y = base_y;
  
  _pointInGridNumX = nx;
  _pointInGridNumY = ny;
  
  // Send back the first point

  return getNextInGrid();
}
  

/**********************************************************************
 * getNextInGrid() - Get the next template grid point within the grid.
 *                   Returns NULL when there are no more points in the
 *                   grid.
 *
 * Returns a pointer to a static object which must NOT be deleted by the
 * calling routine.
 */

GridPoint *GridTemplate::getNextInGrid(void) const
{
  while (_pointInGridIterator != _offsetList.end())
  {
    GridOffset *offset = *_pointInGridIterator;
    
    _pointInGridIterator++;
    
    _pointInGridReturn.x = _pointInGridBase.x + offset->x_offset;
    _pointInGridReturn.y = _pointInGridBase.y + offset->y_offset;

    if (_pointInGridReturn.x >= 0 &&
	_pointInGridReturn.x < _pointInGridNumX &&
	_pointInGridReturn.y >= 0 &&
	_pointInGridReturn.y < _pointInGridNumY)
    {
      return &_pointInGridReturn;
    }
    
  }
  
  return (GridPoint *)NULL;
}
  

/**********************************************************************
 * printOffsetList() - Print the offset list to the given stream.  This
 *                     is used for debugging.
 */

void GridTemplate::printOffsetList(FILE *stream)
{
  vector< GridOffset* >::iterator ol_iterator;
  
  for (ol_iterator = _offsetList.begin();
       ol_iterator != _offsetList.end();
       ol_iterator++)
  {
    GridOffset *offset = *ol_iterator;

    double x = (double)offset->x_offset;
    double y = (double)offset->y_offset;
    
    double distance = sqrt((x * x) + (y * y));
    
    fprintf(stream, " %4d %4d   %f\n",
	    offset->x_offset, offset->y_offset, distance);
    
  }
  
}


/**********************************************************************
 *              Private Member Functions                              *
 **********************************************************************/

/**********************************************************************
 * _addOffset() - Add the given offset to the offset list.
 */

void GridTemplate::_addOffset(int x_offset, int y_offset)
{
  GridOffset *offset = new GridOffset(x_offset, y_offset);
  
  _offsetList.push_back(offset);
  
  return;
}



////////////////////////////////////////////////////////////////
//   GridTemplateFactory
////////////////////////////////////////////////////////////////
GridTemplateFactory::GridTemplateFactory() {
	enum_to_string.resize(GridTemplate_NUM_TEMPLATES);
	
	enum_to_string[GridTemplate_Square] = "SQUARE";
	enum_to_string[GridTemplate_Circle] = "CIRCLE";			
	//enum_to_string[GridTemplate_Rectangle] = "RECTANGLE";
}

GridTemplateFactory::~GridTemplateFactory() {
	enum_to_string.clear();
}


/////////////////////////////////////////////////////////////////
// log & exit on failure
GridTemplateFactory::GridTemplates GridTemplateFactory::string2Enum(string target) {

	for (unsigned int ix = 0; ix < GridTemplate_NUM_TEMPLATES; ix++){		
		if (enum_to_string[ix] == target){
			return static_cast<GridTemplates>(ix);
		}
	}
	mlog << Error << "\nGridTemplateFactory::string2Enum() -> "
		" Unexpected target value of " << target << ".\n\n";	
	exit(1);
}


/////////////////////////////////////////////////////////////////
// log & exit on failure
string GridTemplateFactory::enum2String(GridTemplates target) {

	if( static_cast<int>(target) > enum_to_string.size() - 1){

		mlog << Error << "\nGridTemplateFactory::enum2String() ->"
			" target out of range " << target << " > " << (static_cast<int>(enum_to_string.size()) - 1)
		     << ".\n\n";
		exit(1);
	}
	return enum_to_string[static_cast<int>(target)];
}

/////////////////////////////////////////////////////////////////
// Caller assumes ownership of the returned pointer.
GridTemplate* GridTemplateFactory::buildGT(string gt, int width) {

	return buildGT(string2Enum(gt), width);
}

/////////////////////////////////////////////////////////////////
// Caller assumes ownership of the returned pointer.
GridTemplate* GridTemplateFactory::buildGT(GridTemplates gt, int width) {

	switch (gt) {
	case(GridTemplate_Square):
		return new RectangularTemplate(width,width);

	case(GridTemplate_Circle):
		// circle takes a radius, not a width		
		return new CircularTemplate((width-1)/2);

	default:
		mlog << Error << "\nbuildGT() -> "
			" Unexpected gt value of " << gt << ".\n\n";			
		exit(1);
		
			}
}
